function [MatPat,time] = GRay_Demo(q,num,gnn,gna,para,drawflag)


%2007-8-31
%an matlab implementation for G-Ray
% test on window xp + matlab 7.0.4.365 (R14) Service Pack 2
% requirement: graphviz to show the results (http://www.graphviz.org/)
%
% Inputs:
% q: the query graph
%    q.tyidx: (1 x t) type index, indicate what kind of attribute node we want
%    q.tygph: (t x t) type graph, indicate what kind of relationship we want in the
%              found pattern
%    q.tygph=-1:  another way to say that we want a star structure, with
%                 q.tyidx(1) as core
%    q.sk: 1 x t softand coefficient, not investiaged/implemented yet. you do
%          not need to provide this field
%
% gnn: (n x n) node-node graph;
% gna: (n x m) node-attribute graph
% num: how many patterns do you want
% para: parameter to do rwr
%
%
% Outputs:
%  MatPat: a 1 x num cell matrix, each of its elements is one matching
%  subgraph
%  time: average response time to find one matching subgraph
%
% Paper: this implementation is based on the following paper:
%----------------------------------------------------------------
%        Hanghang Tong, Brian Gallagher, Christos Faloutsos, and Tina Eliassi-Rad
%        Fast Best-Effort Pattern Match in Large Attributed Graphs. 
%        KDD 2007: 737-746
%-----------------------------------------------------------------
% Copyright: 2007
%         Hanghang Tong, Brian Gallagher, 
%         Christos Faloutsos, and Tina Eliassi-Rad         
%         All rights reserved.
% Please address questions or comments to: htong (AT) cs (DOT) cmu (DOT) edu
%----------------------------------------------------------------

% Example to use
%  load ./data/dt_gray_toy.mat  
%  [MatPat,time] = GRay_Demo(q_loop,1,gnn,gna,[10,0.9,2 1 0],[0 1]);
%  find a loop pattern (that match q_loop), using the default the
%  parameter, and then draw the result by graphviz

%whether or not show the original graph and/or the resulting subgraphs
if nargin<6
    drawflag=[0 0];%does not draw anything
end
if nargin<5
    para = [10,0.9,2 1 0];
end

if q.tygph == -1% it is a star-structure
    t = length(q.tyidx);
    tygph = zeros(t,t);
    tygph(1,2:end)=1;
    tygph(2:end,1)=1;
    q.tygph = tygph;
elseif q.tygph ==-2 %it is a line structure
    t = length(q.tyidx);
    tygph = zeros(t,t);
    for i=1:t-1
        tygph(i,i+1)=1;
        tygph(i+1,i)=1;
    end
    q.tygph = tygph;
elseif q.tygph == -3 %loop structure
    
    t = length(q.tyidx);
    tygph = zeros(t,t);
    for i=1:t-1
        tygph(i,i+1)=1;
        tygph(i+1,i)=1;
    end
    tygph(1,t)=1;
    tygph(t,1)=1;
    q.tygph = tygph;
end

tic;
df = 1;%whether or not put a directed edge between 
%node-node-attribute graph
[n,m] = size(gnn);
[n,m] = size(gna);
if df==1
    W = [gnn,sparse(n,m);gna',sparse(m,m)];
else
    W = [gnn,gna;gna',sparse(m,m)];
end

%normalization on W
%get P matrix
D0 = (sparse(diag(sum(W,2))));
for i=1:m+n
    if D0(i,i)>0
        D0(i,i) = 1/D0(i,i);
    end
end
%row normalization
P = D0 * W;
P = P';
P0 = P(1:n,1:n);


%setup Y
%for every query attribute
t = length(q.tyidx);
Y = sparse(n+m,t);
for i=1:t
    Y(q.tyidx(i)+n,i)=1;
end
%individual rwr
[aF]  = PPR_Propagate3(P,para,Y);
aF = aF(1:n,:);

%aF = aF.^5;
%['Attribute Node RWR:' num2str(toc)]
%already visited nodes
vn = [];
for i=1:num%
    [pat,kn,nF] = AppMatOnePat(q,vn,aF,gnn,gna,P0,para);%Approximately match one pattern
    vn = union(vn,kn);
    tmp.pat = pat;
    tmp.kn = kn;
    MatPat{i} = tmp;
    tmp.score = fpm_eval(q,MatPat{i},gnn,gna);
    tmp.score(end+1:end+2) = fpm_eval2(q,tmp,gnn,gna,aF,nF,P0,para);
    MatPat{i} = tmp;
end
time = toc/num;

%draw the original attributed graph 
%(warning: graphviz might `die' if large graph)
if drawflag(1)==1
    DrawAttGraph(gnn,gna);
end
%draw the result
if drawflag(2)~=0
    drawfpm(gnn,gna,MatPat);
end

function [pp,kn,nF] = AppMatOnePat(q,vn,aF,gnn,gna,P0,para);%Approximately match one pattern

%setup iNp,iEp,pp
tidx = q.tyidx;
tg = q.tygph;
t = length(tidx);

%partial matched pattern
pp = [];
iNp = zeros(1,t);%0: un-touched, -1: touched, but not proceeded, 1: proceeded
iEp = tg;
iEp(find(tg~=0))=-1;%un-proceeded

%key node
kn = zeros(1,t);

%setup node-score
[n0,m0] = size(aF);
nF = sparse(n0,m0);

%
[n,n] = size(gnn);
[n,m] = size(gna);

flag=1;
ccnum = 0;
while flag==1
    %Select one attribute node
    an = SelAttNode(tidx,tg,iNp,iEp);
    if an<0%each attribute node has been edited
        flag=-1;
        return;
    end
        
    %Select one promising node,with attribute an
    %at least one new node
    %tic;
    if ccnum==0
        %nn = SelNodeNode(tidx,tg,aF,nF,iNp,iEp,an,gna,vn,kn);
        %Instead we will use SelSeedNode now!
        nn = SelSeedNode(tidx,tg,aF,an,gna,vn,kn);
    else
        nn = SelNodeNode3(tidx,tg,aF,nF,iNp,iEp,an,gna,pp,kn);
    end
    %['Select Node:' num2str(toc)]
    ccnum = ccnum + 1;
    %update iNp
    iNp(an) = 1;%proceeded
    
    %update key node
    kn(an) = nn;
    
    tPos = find(tg(an,:));
    if nn<=0        
        iEp(an,tPos) = 1;
        iEp(tPos,an) = 1;
    else
        %tic;
        %update node-score matrix
        y = sparse(n,1);
        y(nn) = 1;    
        [f]  = PPR_Propagate3(P0,para,y);
        nF(:,an) = f;
        [sf,cnI] = sort(full(f),1,'descend');
        pp = union(pp,nn);
        nn0 = find(cnI==nn);    
        %disp(['Node RWR:' num2str(toc)])


        for i=1:length(tPos);%for each of its adjacent edge
            if iEp(an,tPos(i))==-1%Has not been proceeded
                %Select the ending node,with attribute tPos(i)
                %tic;
                nn2 = SelNodeNode3(tidx,tg,aF,nF,iNp,iEp,tPos(i),gna,pp,kn);
                %['Select Neiboring Node:' num2str(toc)]
                if nn2>0
                    %tic;
                    nn2_0 = find(cnI==nn2);
                    kn(tPos(i)) = (nn2);
                    %EditOndEdge
                    idx = cnI(min(nn0,nn2_0):max(nn0,nn2_0));
                    a = max(min(nn0,nn2_0)-200,1);
                    b = min(max(nn0,nn2_0)+200,length(f));
                    idx2 = cnI(a:b);
                    %to maximize total f, maximum allowed path length 5
                    %blocking the nodes that are already in pp
                    %[pth pflg]= DP_path_Blk(idx,pp,f,gnn,10,1);        
                    %block the nodes that are already in pp
                    %find path by ceps
                    %[pth, pflg] = DP_path_Blk_CePS([idx(1),idx(end)],idx,pp,gnn);
                    
                    %[pth, pflg] = DP_path_Blk_CePS([idx(1),idx(end)],[1:size(gnn,2)],pp,gnn);
                    %[pth, pflg] = Fpm_DP_path_Dij(idx(1),idx(end),pp,f,gnn);
                    [pth, pflg] = Fpm_DP_path_Dij2(idx(1),idx(end),pp,idx2,f,gnn);
                    if pflg>0
                        pp = union(pp,pth);
                    end
                    %disp(['Find Path:' num2str(toc)]);
                end
                %adjacent edges have been proceeded
                iEp(an,tPos(i)) = 1;
                iEp(tPos(i),an) = 1;
            end 
        end       
        
            
        %check for each ajacent node
        for i=1:length(tPos)
            if min(iEp(tPos(i),:))~=-1%each of its adjacent edge has been proceeded
                iNp(tPos(i)) = 1;%node has been proceeded
            else
                iNp(tPos(i)) = -1;%node has been touched, but not proceeded
            end
        end
    end
end
%



function an = SelAttNode(tidx,tg,iNp,iEp);

if min(iNp)==1%every node has been proceeded
    an = -1;
elseif length(find(iNp==0))==length(iNp)%every attribute node has NOT been proceeded
    %find the one with highest degree
    d = sum(tg);
    an = find(d==max(d));
    an = an(1);
else%find the one with highest degree among those has touched
    tmp = find(iNp==-1);
    if length(tmp)>0
        d = sum(tg(tmp,:),2);
        an = find(d==max(d));
        an = tmp(an(1));
    else
        tmp = find(iNp==0);
        d = sum(tg(tmp,:),2);
        an = find(d==max(d));
        an = tmp(an(1));
    end
        
        
end

function nn = SelSeedNode(tidx,tg,aF,an,gna,vn,kn);%find a seed with attribute an


P = fpm_w2p(tg);

%compute all esc_prob, a little lazy here, actually we only need esp_prob
%from an to all others
pe = Prob_Esp_All_Sink(P,0.95);

tPos = setdiff([1:length(tidx)],an);

%score weight
sw = pe(an,tPos);
sw = max(sw,0.0001);
sw = 1./sw;

cp = find(gna(:,tidx(an)));
aF = aF(cp,:);

for i=1:length(tPos)    
    F(:,i) = aF(:,tPos(i)).^(sw(i));    
end

%Here we should insure that we do find a seed, so, possiblly we will issue
%soft-and
nn = -1;
len = length(tPos);
while nn==-1&len>0
    [cF] = Kmore_Prob(F,length(tPos),1);
    [scF,cI] = sort(full(cF),1,'descend');
    cnt = FindFirstNot(cp(cI),vn);
    if cnt>0
        nn = cp(cI(cnt));   
    else
        len = len -1;
    end
end
    






function [cF] = ComScore(F,q)

sk = q.sk;
g = q.tygph;

t = length(sk);
for i=1:t
    pos = find(g(i,:)~=0);
    tF = F(:,union(i,pos));
    cF(:,i) = Kmore_Prob(pV,sk(i),1);
end

function [pos] = CandGen(cF,len)

if nargin<2
    [n,m] = size(cF);
    len = min(400,n);
end

[n,t] = size(cF);
[ccF,I] = sort(full(cF),1,'descend');
pos = [];
for i=1:t
    pos = union(pos,I(1:len,i));
end



function cnt = FindFirstAttNot(I,ref,gna,att)
%find first element in I with attribute att, that is not in ref

flag=1;
cnt=1;
len = length(I);
while flag==1&cnt<=len
    if gna(I(cnt),att)~=0
        if length(find(ref==I(cnt)))==0        
            flag=0;
        else
            cnt = cnt+1;
        end    
    else
        cnt = cnt + 1;
    end
end
if cnt>len
    cnt = -1;
end

function cnt = FindFirstNot(I,ref)
%find first element in I that not in ref

flag=1;
cnt=1;
while flag==1&cnt<=length(I)
    if length(find(ref==I(cnt)))==0        
        flag=0;
    else
        cnt = cnt + 1;
    end
end
if cnt>length(I)
    cnt = -1;
end

function cnt = FindFirstIn(I,ref)
%find first element in I that not in ref

flag=1;
cnt=1;
while flag==1&cnt<=length(I)
    if length(ref==I(cnt))>0        
        flag=0;
    else
        cnt = cnt + 1;
    end
end
if cnt>length(I)
    cnt = -1;
end


function P = fpm_w2p(W)

%normalization on W
%get P matrix
[m,n] = size(W);
D0 = (sparse(diag(sum(W,2))));
for i=1:m
    if D0(i,i)>0
        D0(i,i) = 1/D0(i,i);
    end
end
%row normalization
P = D0 * W;